@@ -9,8 +9,8 @@
  * "browse_callback", and parts of "main".
  * 
  * Everything else is
- * Copyright © 2008-2015 Teddy Hogeborn
- * Copyright © 2008-2015 Björn Påhlsson
+ * Copyright © 2008-2016 Teddy Hogeborn
+ * Copyright © 2008-2016 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -47,7 +47,7 @@
 				   strtof(), abort() */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h>		/* strcmp(), strlen(), strerror(),
-				   asprintf(), strcpy() */
+				   asprintf(), strncpy() */
 #include <sys/ioctl.h>		/* ioctl */
 #include <sys/types.h>		/* socket(), inet_pton(), sockaddr,
 				   sockaddr_in6, PF_INET6,
@@ -57,12 +57,16 @@
 #include <sys/socket.h>		/* socket(), struct sockaddr_in6,
 				   inet_pton(), connect(),
 				   getnameinfo() */
-#include <fcntl.h>		/* open(), unlinkat() */
+#include <fcntl.h>		/* open(), unlinkat(), AT_REMOVEDIR */
 #include <dirent.h>		/* opendir(), struct dirent, readdir()
 				 */
 #include <inttypes.h>		/* PRIu16, PRIdMAX, intmax_t,
 				   strtoimax() */
-#include <errno.h>		/* perror(), errno,
+#include <errno.h>		/* perror(), errno, EINTR, EINVAL,
+				   EAI_SYSTEM, ENETUNREACH,
+				   EHOSTUNREACH, ECONNREFUSED, EPROTO,
+				   EIO, ENOENT, ENXIO, ENOMEM, EISDIR,
+				   ENOTEMPTY,
 				   program_invocation_short_name */
 #include <time.h>		/* nanosleep(), time(), sleep() */
 #include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
@@ -513,7 +517,7 @@ static void debuggnutls(__attribute__((unused)) int level,
   fprintf_plus(stderr, "GnuTLS: %s", string);
 }
 
-__attribute__((nonnull, warn_unused_result))
+__attribute__((nonnull(1, 2, 4), warn_unused_result))
 static int init_gnutls_global(const char *pubkeyfilename,
 			      const char *seckeyfilename,
 			      const char *dhparamsfilename,
@@ -525,13 +529,6 @@ static int init_gnutls_global(const char *pubkeyfilename,
     fprintf_plus(stderr, "Initializing GnuTLS\n");
   }
   
-  ret = gnutls_global_init();
-  if(ret != GNUTLS_E_SUCCESS){
-    fprintf_plus(stderr, "GnuTLS global_init: %s\n",
-		 safer_gnutls_strerror(ret));
-    return -1;
-  }
-  
   if(debug){
     /* "Use a log level over 10 to enable all debugging options."
      * - GnuTLS manual
@@ -545,7 +542,6 @@ static int init_gnutls_global(const char *pubkeyfilename,
   if(ret != GNUTLS_E_SUCCESS){
     fprintf_plus(stderr, "GnuTLS memory error: %s\n",
 		 safer_gnutls_strerror(ret));
-    gnutls_global_deinit();
     return -1;
   }
   
@@ -755,7 +751,6 @@ static int init_gnutls_global(const char *pubkeyfilename,
  globalfail:
   
   gnutls_certificate_free_credentials(mc->cred);
-  gnutls_global_deinit();
   gnutls_dh_params_deinit(mc->dh_params);
   return -1;
 }
@@ -822,54 +817,54 @@ static void empty_log(__attribute__((unused)) AvahiLogLevel level,
 
 /* Set effective uid to 0, return errno */
 __attribute__((warn_unused_result))
-error_t raise_privileges(void){
-  error_t old_errno = errno;
-  error_t ret_errno = 0;
+int raise_privileges(void){
+  int old_errno = errno;
+  int ret = 0;
   if(seteuid(0) == -1){
-    ret_errno = errno;
+    ret = errno;
   }
   errno = old_errno;
-  return ret_errno;
+  return ret;
 }
 
 /* Set effective and real user ID to 0.  Return errno. */
 __attribute__((warn_unused_result))
-error_t raise_privileges_permanently(void){
-  error_t old_errno = errno;
-  error_t ret_errno = raise_privileges();
-  if(ret_errno != 0){
+int raise_privileges_permanently(void){
+  int old_errno = errno;
+  int ret = raise_privileges();
+  if(ret != 0){
     errno = old_errno;
-    return ret_errno;
+    return ret;
   }
   if(setuid(0) == -1){
-    ret_errno = errno;
+    ret = errno;
   }
   errno = old_errno;
-  return ret_errno;
+  return ret;
 }
 
 /* Set effective user ID to unprivileged saved user ID */
 __attribute__((warn_unused_result))
-error_t lower_privileges(void){
-  error_t old_errno = errno;
-  error_t ret_errno = 0;
+int lower_privileges(void){
+  int old_errno = errno;
+  int ret = 0;
   if(seteuid(uid) == -1){
-    ret_errno = errno;
+    ret = errno;
   }
   errno = old_errno;
-  return ret_errno;
+  return ret;
 }
 
 /* Lower privileges permanently */
 __attribute__((warn_unused_result))
-error_t lower_privileges_permanently(void){
-  error_t old_errno = errno;
-  error_t ret_errno = 0;
+int lower_privileges_permanently(void){
+  int old_errno = errno;
+  int ret = 0;
   if(setuid(uid) == -1){
-    ret_errno = errno;
+    ret = errno;
   }
   errno = old_errno;
-  return ret_errno;
+  return ret;
 }
 
 /* Helper function to add_local_route() and delete_local_route() */
@@ -1628,22 +1623,23 @@ static void handle_sigterm(int sig){
 __attribute__((nonnull, warn_unused_result))
 bool get_flags(const char *ifname, struct ifreq *ifr){
   int ret;
-  error_t ret_errno;
+  int old_errno;
   
   int s = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
   if(s < 0){
-    ret_errno = errno;
+    old_errno = errno;
     perror_plus("socket");
-    errno = ret_errno;
+    errno = old_errno;
     return false;
   }
-  strcpy(ifr->ifr_name, ifname);
+  strncpy(ifr->ifr_name, ifname, IF_NAMESIZE);
+  ifr->ifr_name[IF_NAMESIZE-1] = '\0'; /* NUL terminate */
   ret = ioctl(s, SIOCGIFFLAGS, ifr);
   if(ret == -1){
     if(debug){
-      ret_errno = errno;
+      old_errno = errno;
       perror_plus("ioctl SIOCGIFFLAGS");
-      errno = ret_errno;
+      errno = old_errno;
     }
     return false;
   }
@@ -1913,18 +1909,8 @@ void run_network_hooks(const char *mode, const char *interface,
       return;
     }
   }
-#ifdef __GLIBC__
-#if __GLIBC_PREREQ(2, 15)
   int numhooks = scandirat(hookdir_fd, ".", &direntries,
 			   runnable_hook, alphasort);
-#else  /* not __GLIBC_PREREQ(2, 15) */
-  int numhooks = scandir(hookdir, &direntries, runnable_hook,
-			 alphasort);
-#endif	/* not __GLIBC_PREREQ(2, 15) */
-#else	/* not __GLIBC__ */
-  int numhooks = scandir(hookdir, &direntries, runnable_hook,
-			 alphasort);
-#endif	/* not __GLIBC__ */
   if(numhooks == -1){
     perror_plus("scandir");
     return;
@@ -2085,9 +2071,9 @@ void run_network_hooks(const char *mode, const char *interface,
 }
 
 __attribute__((nonnull, warn_unused_result))
-error_t bring_up_interface(const char *const interface,
-			   const float delay){
-  error_t old_errno = errno;
+int bring_up_interface(const char *const interface,
+		       const float delay){
+  int old_errno = errno;
   int ret;
   struct ifreq network;
   unsigned int if_index = if_nametoindex(interface);
@@ -2103,7 +2089,8 @@ error_t bring_up_interface(const char *const interface,
   }
   
   if(not interface_is_up(interface)){
-    error_t ret_errno = 0, ioctl_errno = 0;
+    int ret_errno = 0;
+    int ioctl_errno = 0;
     if(not get_flags(interface, &network)){
       ret_errno = errno;
       fprintf_plus(stderr, "Failed to get flags for interface "
@@ -2212,8 +2199,8 @@ error_t bring_up_interface(const char *const interface,
 }
 
 __attribute__((nonnull, warn_unused_result))
-error_t take_down_interface(const char *const interface){
-  error_t old_errno = errno;
+int take_down_interface(const char *const interface){
+  int old_errno = errno;
   struct ifreq network;
   unsigned int if_index = if_nametoindex(interface);
   if(if_index == 0){
@@ -2222,7 +2209,8 @@ error_t take_down_interface(const char *const interface){
     return ENXIO;
   }
   if(interface_is_up(interface)){
-    error_t ret_errno = 0, ioctl_errno = 0;
+    int ret_errno = 0;
+    int ioctl_errno = 0;
     if(not get_flags(interface, &network) and debug){
       ret_errno = errno;
       fprintf_plus(stderr, "Failed to get flags for interface "
@@ -2478,14 +2466,14 @@ int main(int argc, char *argv[]){
 			 .args_doc = "",
 			 .doc = "Mandos client -- Get and decrypt"
 			 " passwords from a Mandos server" };
-    ret = argp_parse(&argp, argc, argv,
-		     ARGP_IN_ORDER | ARGP_NO_HELP, 0, NULL);
-    switch(ret){
+    ret_errno = argp_parse(&argp, argc, argv,
+			   ARGP_IN_ORDER | ARGP_NO_HELP, 0, NULL);
+    switch(ret_errno){
     case 0:
       break;
     case ENOMEM:
     default:
-      errno = ret;
+      errno = ret_errno;
       perror_plus("argp_parse");
       exitcode = EX_OSERR;
       goto end;
@@ -2500,9 +2488,9 @@ int main(int argc, char *argv[]){
        <http://bugs.debian.org/633582> */
     
     /* Re-raise privileges */
-    ret_errno = raise_privileges();
-    if(ret_errno != 0){
-      errno = ret_errno;
+    ret = raise_privileges();
+    if(ret != 0){
+      errno = ret;
       perror_plus("Failed to raise privileges");
     } else {
       struct stat st;
@@ -2572,9 +2560,9 @@ int main(int argc, char *argv[]){
       }
       
       /* Lower privileges */
-      ret_errno = lower_privileges();
-      if(ret_errno != 0){
-	errno = ret_errno;
+      ret = lower_privileges();
+      if(ret != 0){
+	errno = ret;
 	perror_plus("Failed to lower privileges");
       }
     }
@@ -2908,7 +2896,7 @@ int main(int argc, char *argv[]){
     
     /* Allocate a new server */
     mc.server = avahi_server_new(avahi_simple_poll_get(simple_poll),
-				 &config, NULL, NULL, &ret_errno);
+				 &config, NULL, NULL, &ret);
     
     /* Free the Avahi configuration data */
     avahi_server_config_free(&config);
@@ -2917,7 +2905,7 @@ int main(int argc, char *argv[]){
   /* Check if creating the Avahi server object succeeded */
   if(mc.server == NULL){
     fprintf_plus(stderr, "Failed to create Avahi server: %s\n",
-		 avahi_strerror(ret_errno));
+		 avahi_strerror(ret));
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -2975,7 +2963,6 @@ int main(int argc, char *argv[]){
   
   if(gnutls_initialized){
     gnutls_certificate_free_credentials(mc.cred);
-    gnutls_global_deinit();
     gnutls_dh_params_deinit(mc.dh_params);
   }
   
@@ -3004,9 +2991,9 @@ int main(int argc, char *argv[]){
   
   /* Re-raise privileges */
   {
-    ret_errno = raise_privileges();
-    if(ret_errno != 0){
-      errno = ret_errno;
+    ret = raise_privileges();
+    if(ret != 0){
+      errno = ret;
       perror_plus("Failed to raise privileges");
     } else {
       
@@ -3020,9 +3007,9 @@ int main(int argc, char *argv[]){
 	while((interface=argz_next(interfaces_to_take_down,
 				   interfaces_to_take_down_size,
 				   interface))){
-	  ret_errno = take_down_interface(interface);
-	  if(ret_errno != 0){
-	    errno = ret_errno;
+	  ret = take_down_interface(interface);
+	  if(ret != 0){
+	    errno = ret;
 	    perror_plus("Failed to take down interface");
 	  }
 	}
@@ -3033,9 +3020,9 @@ int main(int argc, char *argv[]){
       }
     }
     
-    ret_errno = lower_privileges_permanently();
-    if(ret_errno != 0){
-      errno = ret_errno;
+    ret = lower_privileges_permanently();
+    if(ret != 0){
+      errno = ret;
       perror_plus("Failed to lower privileges permanently");
     }
   }
@@ -3043,51 +3030,65 @@ int main(int argc, char *argv[]){
   free(interfaces_to_take_down);
   free(interfaces_hooks);
   
-  /* Removes the GPGME temp directory and all files inside */
-  if(tempdir != NULL){
+  void clean_dir_at(int base, const char * const dirname,
+		    uintmax_t level){
     struct dirent **direntries = NULL;
-    int tempdir_fd = (int)TEMP_FAILURE_RETRY(open(tempdir, O_RDONLY
-						  | O_NOFOLLOW
-						  | O_DIRECTORY
-						  | O_PATH));
-    if(tempdir_fd == -1){
+    int dret;
+    int dir_fd = (int)TEMP_FAILURE_RETRY(openat(base, dirname,
+						O_RDONLY
+						| O_NOFOLLOW
+						| O_DIRECTORY
+						| O_PATH));
+    if(dir_fd == -1){
       perror_plus("open");
-    } else {
-#ifdef __GLIBC__
-#if __GLIBC_PREREQ(2, 15)
-      int numentries = scandirat(tempdir_fd, ".", &direntries,
-				 notdotentries, alphasort);
-#else  /* not __GLIBC_PREREQ(2, 15) */
-      int numentries = scandir(tempdir, &direntries, notdotentries,
-			       alphasort);
-#endif	/* not __GLIBC_PREREQ(2, 15) */
-#else	/* not __GLIBC__ */
-      int numentries = scandir(tempdir, &direntries, notdotentries,
-			       alphasort);
-#endif	/* not __GLIBC__ */
-      if(numentries >= 0){
-	for(int i = 0; i < numentries; i++){
-	  ret = unlinkat(tempdir_fd, direntries[i]->d_name, 0);
-	  if(ret == -1){
-	    fprintf_plus(stderr, "unlinkat(open(\"%s\", O_RDONLY),"
-			 " \"%s\", 0): %s\n", tempdir,
+    }
+    int numentries = scandirat(dir_fd, ".", &direntries,
+			       notdotentries, alphasort);
+    if(numentries >= 0){
+      for(int i = 0; i < numentries; i++){
+	if(debug){
+	  fprintf_plus(stderr, "Unlinking \"%s/%s\"\n",
+		       dirname, direntries[i]->d_name);
+	}
+	dret = unlinkat(dir_fd, direntries[i]->d_name, 0);
+	if(dret == -1){
+	  if(errno == EISDIR){
+	      dret = unlinkat(dir_fd, direntries[i]->d_name,
+			      AT_REMOVEDIR);
+	  }	    
+	  if((dret == -1) and (errno == ENOTEMPTY)
+	     and (strcmp(direntries[i]->d_name, "private-keys-v1.d")
+		  == 0) and (level == 0)){
+	    /* Recurse only in this special case */
+	    clean_dir_at(dir_fd, direntries[i]->d_name, level+1);
+	    dret = 0;
+	  }
+	  if(dret == -1){
+	    fprintf_plus(stderr, "unlink(\"%s/%s\"): %s\n", dirname,
 			 direntries[i]->d_name, strerror(errno));
 	  }
-	  free(direntries[i]);
-	}
-	
-	/* need to clean even if 0 because man page doesn't specify */
-	free(direntries);
-	if(numentries == -1){
-	  perror_plus("scandir");
-	}
-	ret = rmdir(tempdir);
-	if(ret == -1 and errno != ENOENT){
-	  perror_plus("rmdir");
 	}
+	free(direntries[i]);
       }
-      close(tempdir_fd);
+      
+      /* need to clean even if 0 because man page doesn't specify */
+      free(direntries);
+      if(numentries == -1){
+	perror_plus("scandirat");
+      }
+      dret = unlinkat(base, dirname, AT_REMOVEDIR);
+      if(dret == -1 and errno != ENOENT){
+	perror_plus("rmdir");
+      }
+    } else {
+      perror_plus("scandirat");
     }
+    close(dir_fd);
+  }
+  
+  /* Removes the GPGME temp directory and all files inside */
+  if(tempdir != NULL){
+    clean_dir_at(-1, tempdir, 0);
   }
   
   if(quit_now){
